--[[ rcon.lua © Penguin_Spy 2024
  This Source Code Form is subject to the terms of the Mozilla Public
  License, v. 2.0. If a copy of the MPL was not distributed with this
  file, You can obtain one at https://mozilla.org/MPL/2.0/.
  This Source Code Form is "Incompatible With Secondary Licenses", as
  defined by the Mozilla Public License, v. 2.0.
 
  The Covered Software may not be used as training or other input data
  for LLMs, generative AI, or other forms of machine learning or neural
  networks.
]]
local socket = require "socket"
local string_pack, string_unpack, packformat = string.pack, string.unpack, "<i4i4i4zx"

-- receive all data in a packet, blocking for plain luasockets
---@return string?, string?
local function receive(s)
    local d, n, r = "", nil, s.receivepartial or s.receive
    while true do
        local _, err, data = r(s, "*a")
        if err ~= "timeout" then return nil, err end  ---@diagnostic disable-line: missing-return-value
        d = d .. data
        local ln = #d
        if not n and ln >= 4 then
            n = string_unpack("<i4", d) + 4 -- length bytes aren't included in length value
        end
        if n and ln >= n then return d end
    end
end

---@class rcon
---@field private sock any
local rcon = {}
local mt = {__index = rcon}

-- sends a command to the server
---@param command string
---@return string|nil # the response from the server, or `nil` on error
---@return string?    # the error message if any
function rcon:send(command)
    self.sock:send(string_pack(packformat, 10 + #command, 3, 2, command))
    local data, err = receive(self.sock)
    if not data then return nil, err end
    local _, id, _, response = string_unpack(packformat, data)  ---@cast response string
    if id ~= 3 then return nil, "send failed" end
    return response
end

-- closes the rcon connection
function rcon:close() self.sock:close() end

-- authenticates on the provided socket, useful for working with `copas.wrap()`ed sockets:
---```lua
--- rcon.login(copas.wrap(socket.connect("localhost", 25575)), "password")
---```
--- see `rcon.connect()` for the simpler form
---@param s any the socket
---@param password string
---@return rcon|nil # the rcon object, or `nil` on error
---@return string?    # the error message if any
local function login(s, password)
    s:send(string_pack(packformat, 10 + #password, 3, 3, tostring(password)))
    local data, err = receive(s)
    if not data then s:close() return nil, err end
    local _, id = string_unpack(packformat, data)
    if id ~= 3 then s:close() return nil, "authentication failed" end
    return setmetatable({sock = s}, mt) ---@diagnostic disable-line: return-type-mismatch
end

-- creates a rcon connection to the server \
---see `rcon.login()` for working with copas
---@param address string    the address of the server, for `socket.connect()`
---@param port integer      the rcon port
---@param password string   the rcon password
---@return rcon|nil     # the rcon object, or `nil` on error
---@return string?        # the error message if any
local function connect(address, port, password)
    local s, err = socket.connect(address, port)
    if not s then return nil, err end
    s:settimeout(0)
    return login(s, password)
end

return { connect = connect, login = login }